part of masamune.local;

/// Manage Local documents.
/// 
/// Basically load, set the value if there is a value to update and [save()].
/// 
/// ```
/// LocalDocument doc = LocalDocument.load( "user/user" );
/// String name = doc.getString( "name" );
/// doc["age"] = 18;
/// doc.save();
/// ```
class LocalDocument
  extends Document<DataField>
  implements IDataDocument<DataField> {
  /// Process to create a new instance.
  /// 
  /// Do not use from outside the class.
  /// 
  /// [path]: Destination path.
  /// [isTemporary]: True if the data is temporary.
  @override
  @protected
  T createInstance<T extends IClonable>( String path, bool isTemporary ) => LocalDocument._(
    path: path,
    isTemporary: isTemporary,
    group: this.group,
    order: this.order
  ) as T;

  /// Manage Local documents.
  /// 
  /// Basically load, set the value if there is a value to update and [save()].
  /// 
  /// ```
  /// LocalDocument doc = LocalDocument.load( "user/user" );
  /// String name = doc.getString( "name" );
  /// doc["age"] = 18;
  /// doc.save();
  /// ```
  /// 
  /// Request and listen to a document.
  /// 
  /// [path]: Document path.
  static LocalDocument load( String path ){  
    path = path?.applyTags();
    if( isEmpty(path) ) {
      Log.error( "Path is invalid." );
      return null;
    }
    int length = Paths.length( path );
    if( length <= 0 || length % 2 != 0 ) {
      Log.error( "Path is not document path." );
      return null;
    }
    LocalDocument document = PathMap.get<LocalDocument>( path );
    if( document != null ) {      
      return document.reload( );
    }
    document = LocalDocument._( path:path );
    document._loadFromPrefs( );
    return document;
  }
  /// Manage Local documents.
  /// 
  /// Basically load, set the value if there is a value to update and [save()].
  /// 
  /// ```
  /// LocalDocument doc = LocalDocument.load( "user/user" );
  /// String name = doc.getString( "name" );
  /// doc["age"] = 18;
  /// doc.save();
  /// ```
  /// 
  /// Create a LocalDocument that has not yet been synchronized with the local database.
  /// 
  /// [path]: Document path
  /// [data]: Data set locally
  static LocalDocument create( String path, [Map<String, dynamic> data]){
    path = path?.applyTags();
    if( isEmpty(path) ) {
      Log.error( "Path is invalid." );
      return null;
    }
    int length = Paths.length( path );
    if( length <= 0 || length % 2 != 0 ) {
      Log.error( "Path is not document path." );
      return null;
    }
    LocalDocument document = PathMap.get<LocalDocument>( path );
    if( document != null ){
      if( data != null ) document.set( _convertData( path, data ) );
      return document;
    }
    document = LocalDocument._( path:path, isTemporary: true );
    if( data != null ) document.set( _convertData( path, data ) );
    return document;
  }
  /// Manage Local documents.
  /// 
  /// Basically load, set the value if there is a value to update and [save()].
  /// 
  /// ```
  /// LocalDocument doc = LocalDocument.load( "user/user" );
  /// String name = doc.getString( "name" );
  /// doc["age"] = 18;
  /// doc.save();
  /// ```
  /// 
  /// Delete the document at the specified path.
  /// 
  /// [path]: Document path to delete.
  static void deleteAt( String path ) async {
    path = path?.replaceAll("https", "firestore")?.applyTags();
    if( isEmpty(path) ) {
      Log.error( "Path is invalid." );
      return;
    }
    int length = Paths.length( path );
    if( length <= 0 || length % 2 != 0 ) {
      Log.error( "Path is not document path." );
      return;
    }
    LocalDocument document = PathMap.get<LocalDocument>( path );
    if( document == null ) document = LocalDocument._( path:path );
    document.delete( );
  }
  LocalDocument._(
    {
      String path,
      Iterable<DataField> children,
      bool isTemporary = false,
      int group = 0,
      int order = 10
    }
  ) : super(
    path,
    children:children,
    isTemporary:isTemporary, 
    group:group,
    order:order
  );
  static Iterable<DataField> _convertData(
    String path,
    Map<String, dynamic> data
  ) {
    List<DataField> list = ListPool.get();
    data?.forEach(
      (key, value){
        if( isEmpty(key) || value == null ) return;
        list.add( DataField( Paths.child( path, key ), value ) );
      }
    );
    return list;
  }

  /// Update document data.
  LocalDocument reload( ){
    this._loadFromPrefs();
    return this;
  }  
  void _loadFromPrefs( ) {
    try { 
      String json = Prefs.getString("local://" + this.path);
      if( isEmpty( json ) ) return;
      Map<String, dynamic> map = Json.decodeAsMap(json);
      this._setInternal(map);
      this.notifyUpdate();
    } catch( e ) {
      Log.error( e.toString() );
    }
  }
  void _setInternal( Map<String, dynamic> data ) {
      if( data == null ) return;
      data.forEach(
        (key, value){
          if( isEmpty(key) || value == null ) return;
          this[key] = value;
        }
      );
      List<String> list = List.from(this.data.keys);
      for( String tmp in list ) {
        if( !data.containsKey( tmp ) ) this.remove( tmp );
      }
      list.release();
      Log.ast( "Updated data: %s (%s)", [this.path, this.runtimeType] );
  }

  /// Apply the changed document data to the local database.
  Future<T> save<T extends IDataDocument>( ) {
    if( this.isDisposed ) return Future.delayed(Duration.zero);
    if( this.data == null || this.data.length <= 0 ) {
      Prefs.remove( "local://" + this.path );
    } else {
      String json = Json.encode(this.data.keys);
      if( isEmpty( json ) ){
        Prefs.remove( "local://" + this.path );
      } else {
        Prefs.set( "local://" + this.path, json);
      }
    }
    LocalCollection collection = LocalCollection.load( Paths.parent( this.path ) );
    if( collection != null ) collection._saveInternal( this );
    this.notifyUpdate( );
    return this.asFuture();
  }
  /// Delete this FirestoreDocument path from the local database.
  /// 
  /// This object itself is also [dispose()].
  Future delete( ) {
    if( this.isDisposed ) return Future.delayed(Duration.zero);
    Prefs.remove( "local://" + this.path );
    LocalCollection collection = LocalCollection.load( Paths.parent( this.path ) );
    if( collection != null ) collection._deleteInternal( this );
    this.dispose();
    return Future.delayed(Duration.zero);
  }


  /// Get the protocol of the path.
  @override
  String get protocol => Protocol.local;


  /// Get the data.
  /// 
  /// Do not use from external class.
  /// 
  /// When using from an external class, use getInt or getString.
  /// 
  /// [key]: Key for retrieving data.
  @protected
  dynamic operator []( String key ){
    if( !this.data.containsKey( key ) ) return null;
    return this.data[key]?.data;
  }
  /// Set the data
  /// 
  /// [key]: Key for storing data [value]: Data to store.
  void operator []=( String key, dynamic value ){
    if( value == null ) { return; }
    if( this.isLock ) { Log.error( "Data modification is prohibited." ); return; }
    if( value is DataField ){
      this.set( [value] );
    } else {
      if( PathFilter.setTemporary( Paths.child(this.path, key), value) ) return;
      if( this.data.containsKey( key ) && this.data[key] is IUnit<dynamic> ) {
        this.data[key].data = value;
      } else {
        Path path = this.rawPath.clone( );
        path.path = Paths.child( this.rawPath.path, key );
        this.set( [DataField( path.path, value )] );
      }
    }
  }

  /// Gets value as Bool.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  bool getBool(String key, [bool defaultValue = false]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getBool( defaultValue );
  }
  /// Gets value as Double.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  double getDouble(String key, [double defaultValue = 0]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getDouble( defaultValue );
  }
  /// Gets value as Int.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  int getInt(String key, [int defaultValue = 0]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getInt( defaultValue );
  }
  /// Gets value as List<T>.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  List<T> getList<T extends Object>(
    String key,
    [
      List<T> defaultValue = const []
    ]
  ) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getList<T>( defaultValue );
  }
  /// Gets value as Map<K,V>.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  Map<K, V> getMap<K extends Object, V extends Object>(
    String key,
    [
      Map<K, V> defaultValue = const {}
    ]
  ) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getMap<K,V>( defaultValue );
  }
  /// Gets value as String.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  String getString(String key, [String defaultValue = Const.empty]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    DataField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getString( defaultValue );
  }


  /// Get the UID of the document.
  /// 
  /// If there is no value in the field, id will be output.
  String get uid => this.getString(Const.uid, this.id);
  /// Get time.
  /// 
  /// UpdatedTime is output if the field has no value.
  int get time => this.getInt(Const.time, this.updatedTime);

}